/****************************************************************************
**  CUBE        http://www.score-p.org/                                    **
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2015                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  Copyright (c) 2009-2015                                                **
**  German Research School for Simulation Sciences GmbH,                   **
**  Laboratory for Parallel Programming                                    **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/

/**
 * \file
 * \brief Defines a metric class, which uses build in type as a data type
 *
 */
#ifndef CUBE_INCLUSIVE_BUILDIN_METRIC_H
#define CUBE_INCLUSIVE_BUILDIN_METRIC_H


#include <iosfwd>
#include <map>
#include <cstring>
#include <limits>

#include "CubeIDdeliverer.h"
#include "CubeWideSearchEnumerator.h"

#include "CubeMetricBuildInType.h"
#include "CubeServices.h"

/*
   *----------------------------------------------------------------------------
 *
 * class BuildInTypeMetric
 *
   =======
 */


namespace cube
{
template <class T>
class InclusiveBuildInTypeMetric : public BuildInTypeMetric<T>
{
public:

    InclusiveBuildInTypeMetric<T>( const std::string & disp_name,
                                   const std::string & uniq_name,
                                   const std::string & dtype,
                                   const std::string & uom,
                                   const std::string & val,
                                   const std::string & url,
                                   const std::string & descr,
                                   FileFinder *        ffinder,
                                   Metric *            parent,
                                   uint32_t id = 0,
                                   VizTypeOfMetric is_ghost = CUBE_METRIC_NORMAL
                                   )
        :
          BuildInTypeMetric<T>( disp_name,
                                uniq_name,
                                dtype,
                                uom,
                                val,
                                url,
                                descr,
                                ffinder,
                                parent,
                                id,
                                is_ghost
                                )
    {
    }

    virtual
    ~InclusiveBuildInTypeMetric<T>()
    {
    }

    virtual TypeOfMetric
    get_type_of_metric() const
    {
        return CUBE_METRIC_INCLUSIVE;
    }

    virtual IndexFormat
    get_index_format()
    {
        return CUBE_INDEX_FORMAT_DENSE;
    }

    virtual std::string
    get_metric_kind() const
    {
        return "INCLUSIVE";
    };

    virtual row_of_objects_t*
    create_calltree_id_maps( IDdeliverer*,
                             Cnode*,
                             row_of_objects_t*  );                               // /< Creates enumeration of Callpath-Tree according the "deep search" prescription. It is optimal for the calculation of inclusive values in exclusive metric.


//    using cube::BuildInTypeMetric<T>::get_sev;

    // /  Calculates inclusive value for given thread. Expands only calltree. Returns double representation of Value if possible.
    virtual double
    get_sev( Cnode*             cnode,
             CalculationFlavour cfv,
             Sysres*            sys,
             CalculationFlavour sf );


    // / Calculates inclusive value for given cnode and over all threads. Sums up over all threads. Assumed that all threads have continueos ids. Expands only calltree. Returns double representation of Value if possible.
    virtual double
    get_sev( Cnode*             cnode,
             CalculationFlavour cfv );


    virtual char*
    get_sevs_raw( Cnode*             cnode,
                  CalculationFlavour cfv
                  );
};




template <class T>
row_of_objects_t*
InclusiveBuildInTypeMetric<T>::create_calltree_id_maps( IDdeliverer* ids, Cnode* root, row_of_objects_t* _row  )
{
    WideSearchEnumerator enumerator;
    ids->reset();
    _row = enumerator.get_objects_to_enumerate( root, _row );

    for ( row_of_objects_t::iterator iter = _row->begin(); iter < _row->end(); ++iter )
    {
        if ( this->calltree_local_ids.size() <= ( *iter )->get_id() )
        {
            this->calltree_local_ids.resize( ( *iter )->get_id() + 1  );
        }
        this->calltree_local_ids[ ( *iter )->get_id() ] = ids->get_next_id();
    }

    return _row;
}




// / ------------------------ Sum over all threads for given cnode and all subcnodes


template <class T>
double
InclusiveBuildInTypeMetric<T>::get_sev( Cnode* cnode, CalculationFlavour cnf )
{
    if ( !( this->active ) )
    {
        return 0.;
    }


    T v = static_cast<T>( 0 );
    if ( ( this->t_cache )->isTCachedValue( cnode, cnf )  )
    {
        return ( this->t_cache )->getTCachedValue( cnode, cnf );
    }

    if ( this->adv_sev_mat == NULL )
    {
        return v;
    }

    if ( !cnode->is_clustered() )
    {
        v = ( this->adv_sev_mat )->template sumBuildInTypeRow<T>( this->calltree_local_ids[ cnode->get_remapping_cnode()->get_id() ] );
    }
    else
    {
        size_t sysv_size = this->sysv.size();
        for ( size_t i = 0; i < sysv_size; i++ )
        {
            Thread*  _thrd        = this->sysv[ i ];
            Process* _proc        = _thrd->get_parent();
            int64_t  process_rank = _proc->get_rank();
            T        tmp          = ( this->adv_sev_mat )->template getBuildInValue<T>( this->calltree_local_ids[ cnode->get_remapping_cnode( process_rank  )->get_id() ], _thrd->get_id() );
            int64_t  _norm        = cnode->get_cluster_normalization( process_rank );
            if ( _norm > 0 )
            {
                tmp /= ( ( uint64_t )_norm );
            }
            ( v ) += tmp;
        }
    }




    if ( cnf == CUBE_CALCULATE_EXCLUSIVE &&  ( cnode->num_children() > 0 ) )
    {
        std::vector<Cnode*> _cnodes;  // collection of cnodes for the calculation
        for ( cnode_id_t cid = 0; cid < cnode->num_children(); cid++  )
        {
            Cnode* __c  = cnode->get_child( cid );
            bool   _add = ( !__c->isHidden() );
            if ( _add )
            {
                _cnodes.push_back( __c );
            }
        }

        for ( std::vector<Cnode*>::const_iterator _iter = _cnodes.begin(); _iter != _cnodes.end(); ++_iter )
        {
            Cnode* __c = *_iter;
            if ( !__c->is_clustered() )
            {
                v -= ( this->adv_sev_mat )->template sumBuildInTypeRow<T>( this->calltree_local_ids[ __c->get_remapping_cnode()->get_id() ] );
            }
            else
            {
                size_t sysv_size = this->sysv.size();
                for ( size_t i = 0; i < sysv_size; i++ )
                {
                    Thread*  _thrd        = this->sysv[ i ];
                    Process* _proc        = _thrd->get_parent();
                    int64_t  process_rank = _proc->get_rank();
                    T        tmp          = ( this->adv_sev_mat )->template getBuildInValue<T>( this->calltree_local_ids[ __c->get_remapping_cnode( process_rank  )->get_id() ], _thrd->get_id() );
                    int64_t  _norm        = __c->get_cluster_normalization( process_rank );
                    if ( _norm > 0 )
                    {
                        tmp /= ( ( uint64_t )_norm );
                    }
                    ( v ) -= tmp;
                }
            }
        }
    }
    this->t_cache->setTCachedValue( v,  cnode, cnf );
    return static_cast<double>( v );
}








template <class T>
double
InclusiveBuildInTypeMetric<T>::get_sev( Cnode* cnode, CalculationFlavour cnf, Sysres* sys, CalculationFlavour sf )
{
    if ( !( this->active ) )
    {
        return 0.;
    }

    T v = static_cast<T>( 0 );
    if ( ( this->adv_sev_mat == NULL ) || (  ( sys->isSystemTreeNode() || sys->isLocationGroup() ) && ( sf == cube::CUBE_CALCULATE_EXCLUSIVE ) ) )
    {
        return 0.;
    }

    if ( ( this->t_cache )->isTCachedValue( cnode, cnf, sys, sf )  )
    {
        return ( this->t_cache )->getTCachedValue( cnode, cnf, sys, sf );
    }


    if ( sys->isSystemTreeNode() )
    {
        SystemTreeNode* _sys = ( SystemTreeNode* )sys;
        // first add values of all sub system nodes
        for ( unsigned i = 0; i < _sys->num_children(); i++ )
        {
            T _v = this->get_sev( cnode, cnf, _sys->get_child( i ), cube::CUBE_CALCULATE_INCLUSIVE );
            v += _v;
        }
        // then add all values of all sub local groups
        for ( unsigned i = 0; i < _sys->num_groups(); i++ )
        {
            T _v = this->get_sev( cnode, cnf, _sys->get_location_group( i ), cube::CUBE_CALCULATE_INCLUSIVE );
            ( v ) += _v;
        }
    }
    if ( sys->isLocationGroup() )
    {
        LocationGroup* _lg = ( LocationGroup* )sys;

        for ( unsigned i = 0; i < _lg->num_children(); i++ )
        {
            T _v = get_sev( cnode, cnf, _lg->get_child( i ), cube::CUBE_CALCULATE_INCLUSIVE );
            ( v ) += _v;
        }
    }
    if ( sys->isLocation() )
    {
        Location*      _loc    = ( Location* )sys;
        LocationGroup* _lg     = _loc->get_parent();
        int64_t        lg_rank = _lg->get_rank();
        T              _v      = ( this->adv_sev_mat )->template getBuildInValue<T>( this->calltree_local_ids[ cnode->get_remapping_cnode( lg_rank  )->get_id() ], _loc->get_id() );
        int64_t        _norm   = cnode->get_cluster_normalization( lg_rank );
        if ( _norm > 0 )
        {
            _v /= ( ( uint64_t )_norm );
        }

        if ( cnf == CUBE_CALCULATE_EXCLUSIVE &&  ( cnode->num_children() > 0 ) )
        {
            std::vector<Cnode*> _cnodes;  // collection of cnodes for the calculation
            for ( cnode_id_t cid = 0; cid < cnode->num_children(); cid++  )
            {
                Cnode* __c  = cnode->get_child( cid );
                bool   _add = ( !__c->isHidden() );
                if ( _add )
                {
                    _cnodes.push_back( __c );
                }
            }
            for ( std::vector<Cnode*>::const_iterator _iter = _cnodes.begin(); _iter != _cnodes.end(); ++_iter )
            {
                Cnode*  __c   = *_iter;
                T       __v   = ( this->adv_sev_mat )->template getBuildInValue<T>( this->calltree_local_ids[ __c->get_remapping_cnode( lg_rank  )->get_id() ], _loc->get_id() );
                int64_t _norm = __c->get_cluster_normalization( lg_rank );
                if ( _norm > 0 )
                {
                    __v /= ( ( uint64_t )_norm );
                }
                _v -= __v;
            }
        }

        ( v ) += _v;
    }
    this->t_cache->setTCachedValue( v, cnode, cnf, sys, sf );
    return static_cast<double>( v );
}




template <class T>
char*
InclusiveBuildInTypeMetric<T>::get_sevs_raw( Cnode* cnode, CalculationFlavour cnf )
{
    if ( !( this->active ) )
    {
        return NULL;
    }


    if ( this->adv_sev_mat == NULL )
    {
        return NULL;
    }

    char* v = ( this->t_cache )->getCachedRow( cnode, cnf );
    if ( v != NULL )
    {
        return v;
    }

    size_t sysv_size  = this->sysv.size();
    char*  to_return  = services::create_raw_row( sysv_size * this->metric_value->getSize() );
    T*     _to_return = ( T* )to_return;

    if ( !cnode->is_clustered() )
    {
        char* _tmp =    this->adv_sev_mat->getRow( this->calltree_local_ids[ cnode->get_remapping_cnode()->get_id() ] );
        if ( _tmp != NULL )
        {
            memcpy( to_return, _tmp, sysv_size * this->metric_value->getSize() );
        }
    }
    else
    {
        for ( size_t i = 0; i < sysv_size; i++ )
        {
            Thread*  _thrd        = this->sysv[ i ];
            Process* _proc        = _thrd->get_parent();
            int64_t  process_rank = _proc->get_rank();
            T        tmp          = ( this->adv_sev_mat )->template getBuildInValue<T>( this->calltree_local_ids[ cnode->get_remapping_cnode( process_rank  )->get_id() ], _thrd->get_id() );
            int64_t  _norm        = cnode->get_cluster_normalization( process_rank );
            if ( _norm > 0 )
            {
                tmp /= ( ( uint64_t )_norm );
            }
            _to_return[ i ] = tmp;
        }
    }




    if ( cnf == CUBE_CALCULATE_EXCLUSIVE &&  ( cnode->num_children() > 0 ) )
    {
        std::vector<Cnode*> _cnodes;  // collection of cnodes for the calculation
        for ( cnode_id_t cid = 0; cid < cnode->num_children(); cid++  )
        {
            Cnode* __c  = cnode->get_child( cid );
            bool   _add = ( !__c->isHidden() );
            if ( _add )
            {
                _cnodes.push_back( __c );
            }
        }

        for ( std::vector<Cnode*>::const_iterator _iter = _cnodes.begin(); _iter != _cnodes.end(); ++_iter )
        {
            Cnode* __c = *_iter;


            if ( !__c->is_clustered() )
            {
                char* _tmp =    this->adv_sev_mat->getRow( this->calltree_local_ids[ __c->get_remapping_cnode()->get_id() ] );
                T*    tmp  = ( T* )_tmp;

                if ( tmp != NULL )
                {
                    for ( size_t i = 0; i < sysv_size; i++ )
                    {
                        _to_return[ i ] -= tmp[ i ];
                    }
                }
            }
            else
            {
                size_t sysv_size = this->sysv.size();
                for ( size_t i = 0; i < sysv_size; i++ )
                {
                    Thread*  _thrd        = this->sysv[ i ];
                    Process* _proc        = _thrd->get_parent();
                    int64_t  process_rank = _proc->get_rank();
                    T        tmp          = ( this->adv_sev_mat )->template getBuildInValue<T>( this->calltree_local_ids[ __c->get_remapping_cnode( process_rank  )->get_id() ], _thrd->get_id() );
                    int64_t  _norm        = __c->get_cluster_normalization( process_rank );
                    if ( _norm > 0 )
                    {
                        tmp /= ( ( uint64_t )_norm );
                    }
                    _to_return[ i ] -= tmp;
                }
            }
        }
    }
    ( this->t_cache )->setCachedRow( to_return,  cnode, cnf );

    return to_return;
}
}


#endif
